/*****************************************************************************
ETHAN 1.0 

Automated detection of spherical particles from electron micrographs
Copyright (C) 2000 Teemu Kivioja (Teemu.Kivioja@cs.Helsinki.FI)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.
****************************************************************************/


#include "ethan.h"

#define  SQRT_2           1.4142136
#define  MAX_LINE_LENGTH  100

/* program parameters, the default values are set and possible            */
/* configuration file is read in read_conf                                */
float height_param;
float square_width_param;
float ring_width_param;
float dist_param;
float count_param;

int sector_test;
int height_test;
int distance_test;
int refinement;
int make_result_image;

/* some other global variables which have one value throughout the program */
int   size;              /* size of one side of virus image */
int   file_number;       /* starting number of virus file numbering */
int   border;            /* height of margin area needed in filtering */ 
float radius_coff;       /* how much the radius of ring filters differs */

int number(char *s);
int read_conf(char *filename);



/********/
/* main */
/********/
int main(int argc, char *argv[])
{
  int r;
  char *dir_name;
  int sample_rate;

  /* do not produce virus image files */
  if ((argc == 4 || argc == 5) && number(argv[1]) && (atoi(argv[1]) > 0)) {
      dir_name = NULL;
      r = atoi(argv[1]);   
   
      if (argc == 5) 
          sample_rate = read_conf(argv[4]);
      else
          sample_rate = read_conf(NULL);      
                
      find_viruses(r, argv[2], argv[3], dir_name, 
                   sample_rate);
  /* produce virus files */
  } else if ((argc == 7 || argc == 8) && number(argv[1]) && number(argv[5]) && 
              number(argv[6]) && (atoi(argv[1]) > 0) && (atoi(argv[5]) > 0) && 
	      (atoi(argv[6]) >= 0)) {    
      r = atoi(argv[1]);
   
      if (argc == 8) 
          sample_rate = read_conf(argv[7]);
      else
          sample_rate = read_conf(NULL);
                
      dir_name = argv[4];
      size = atoi(argv[5]);
      file_number = atoi(argv[6]); 

      find_viruses(r, argv[2], argv[3], dir_name, sample_rate);

  /* wrong arguments or just printing the arguments */
  } else {
      fprintf(stderr, "ETHAN 1.0\n\n");                            
      fprintf(stderr, "Usage: ethan RADIUS SOURCE DEST [DIR SIZE ");
      fprintf(stderr, "STARTNUMBER] [PATH]\n\n");
      fprintf(stderr, "Find viruses from mrc file SOURCE.\nThe viruses ");
      fprintf(stderr, "to be found should have radius RADIUS in pixels.\n");
      fprintf(stderr, "Coordinates of the viruses are stored in file ");
      fprintf(stderr, "DEST.\n\n");
      fprintf(stderr, "Optionally viruses can be cut out and stored in ");
      fprintf(stderr, "mrc files in the directory DIR.\n");
      fprintf(stderr, "Then SIZE is the length of one side of the virus ");
      fprintf(stderr, "file and files are named\np_NUMBER.mrc. ");
      fprintf(stderr, "The first NUMBER is STARTNUMBER.\n\n");
      fprintf(stderr, "The last optional argument PATH is the path of ");
      fprintf(stderr, "a configuration file.\n");
      return(EXIT_SUCCESS);
  }

  return(EXIT_SUCCESS);
}



/**************************************************************************/
/* Check that string s represents a positive integer                      */
/**************************************************************************/
int number(char *s)
{
    int n, i;
  

    n = strlen(s);
    
    for (i = 0; i < n; i++)
        if (!isdigit(s[i]))
            return NO;
    return YES;

}



/*************************************************************************/
/* set parameters                                                        */
/*************************************************************************/
int read_conf(char *filename)
{
    FILE *fp;
    char line[MAX_LINE_LENGTH];
    char *start;
    int sample_rate;

    /* default values, explanations in configuration file */
    height_param       = 0.5;
    square_width_param = 1;       
    ring_width_param   = 1.2; 
    dist_param         = 0.1;
    sample_rate        = 0;
    radius_coff        = 0.05;

    make_result_image  = 0;

    sector_test        = 1;
    height_test        = 1;
    distance_test      = 1; 
    refinement         = 1;


    /* if no configuration file is given, default values are used */
    if (filename == NULL)
        return sample_rate;    

    /* read values from configuration file, '#' starts a comment line */
    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Ethan: error reading file %s!\n", filename );
	exit(ETHAN_READ_ERROR);    
    }
    
    while ((fgets(line, MAX_LINE_LENGTH, fp)) != NULL) {
        if (!((strcmp(line, "") == 0) || (line[0] == '#'))) {

            if ((start = strstr(line, "HEIGHT_PARAM")) != NULL) {
                start += strlen("HEIGHT_PARAM");
                if (sscanf(start, "%f", &height_param) != 1 || 
                    height_param < 0 || height_param > 1) {
                    fprintf(stderr, 
                            "Ethan: error reading height_param from %s\n", 
                             filename);
                    exit(ETHAN_READ_ERROR);
		}
            } 
            
            if ((start = strstr(line, "SQUARE_WIDTH_PARAM")) != NULL) {
                start += strlen("SQUARE_WIDTH_PARAM");
                if (sscanf(start, "%f", &square_width_param) != 1 || 
                    square_width_param < 0) {
                    fprintf(stderr, 
                       "Ethan: error reading square_width_param from %s\n",
                       filename);
                    exit(ETHAN_READ_ERROR);
		}
            }

             
            if ((start = strstr(line, "RING_WIDTH_PARAM")) != NULL) {
                start += strlen("RINGH_WIDTH_PARAM");
                if (sscanf(start, "%f", &ring_width_param) != 1 || 
                    ring_width_param < 0) {
                    fprintf(stderr, 
                            "Ethan: error reading ring_width_param from %s\n", 
                             filename);
                    exit(ETHAN_READ_ERROR);
		}
            }

            if ((start = strstr(line, "DIST_PARAM")) != NULL) {
                start += strlen("DIST_PARAM");
                if (sscanf(start, "%f", &dist_param) != 1 || 
                    dist_param < 0) {
                    fprintf(stderr, 
                            "Ethan: error reading dist_param from %s\n", 
                             filename);
                    exit(ETHAN_READ_ERROR);
		}
            }

            if ((start = strstr(line, "REDUCTION_PARAM")) != NULL) {
                start += strlen("REDUCTION_PARAM");
                if (sscanf(start, "%d", &sample_rate) != 1 || 
                    sample_rate < 0) {
                    fprintf(stderr, 
                            "Ethan: error reading sample_param from %s\n", 
                             filename);
                    exit(ETHAN_READ_ERROR);
		}
            }


            if ((start = strstr(line, "MAKE_RESULT_IMAGE")) != NULL) {
                start += strlen("MAKE_RESULT_IMAGE");
                if (sscanf(start, "%1d", &make_result_image) != 1 || 
                    (make_result_image != 0 && make_result_image != 1)) {
                    fprintf(stderr, 
                            "Error reading make_result_image from %s\n", 
                             filename);
                    exit(EXIT_FAILURE);
		}
            }


            if ((start = strstr(line, "REFINEMENT_PARAM")) != NULL) {
                start += strlen("REFINEMENT_PARAM");
                if (sscanf(start, "%1d", &refinement) != 1 || 
                    (refinement != 0 && refinement != 1)) {
                    fprintf(stderr, 
                            "Error reading refinement_param from %s\n", 
                             filename);
                    exit(EXIT_FAILURE);
		}
            }


            if ((start = strstr(line, "SECTOR_TEST")) != NULL) {
                start += strlen("SECTOR_TEST");
                if (sscanf(start, "%1d", &sector_test) != 1 || 
                    (sector_test != 0 && sector_test != 1)) {
                    fprintf(stderr, 
                            "Error reading sector_test from %s\n", 
                             filename);
                    exit(EXIT_FAILURE);
		}
            }


            if ((start = strstr(line, "HEIGHT_TEST")) != NULL) {
                start += strlen("HEIGHT_TEST");
                if (sscanf(start, "%1d", &height_test) != 1 || 
                    (height_test != 0 && height_test != 1)) {
                    fprintf(stderr, 
                            "Error reading height_test from %s\n", 
                             filename);
                    exit(EXIT_FAILURE);
		}
            }


            if ((start = strstr(line, "DISTANCE_TEST")) != NULL) {
                start += strlen("DISTANCE_TEST");
                if (sscanf(start, "%1d", &distance_test) != 1 || 
                    (distance_test != 0 && distance_test != 1)) {
                    fprintf(stderr, 
                            "Error reading distance_test from %s\n", 
                             filename);
                    exit(EXIT_FAILURE);
		}
            }
                           
        }
    }
    fclose(fp);
    return sample_rate;
}



/**************************************************************************/
/* set the various values and allocate memory related to image            */
/* to be processed                                                        */
/**************************************************************************/
void allocate_images(int r, int width, int height, int item_size, 
                     int *sampled_width, int *sampled_height, int *s_rate, 
                     int *rb, int *rs, int *square_width, int *n_squares_x, 
                     int *n_squares_y, void ***image, void ***filtered_image, 
                     void ***circle_image, void **rowp, struct peak **peaks)
{
    int piece_height;
    int rbt[5], rst[5];     
  
    /* set sample rate automatically if not set in the configuration file */
    if (*s_rate == 0) {
        *s_rate = r / 25;
        if (*s_rate == 0)
            *s_rate = 1;
    }

    printf("Averaging %d pixels\n", 
           (*s_rate) * (*s_rate));


    /* count the sampled values for radius and square_width */
    *rs = r / *s_rate;

    *rb = ring_width_param * (r / *s_rate) + 1;
    if (*rb - *rs < 3)
       *rb = *rs + 3;

    *square_width = square_width_param * (*rs / SQRT_2);  


    if ((width % *s_rate) == 0)
        *sampled_width = width / *s_rate;                     
    else                                                  
        *sampled_width = (width / *s_rate) + 1;

    if ((height % *s_rate) == 0)
        *sampled_height = height / *s_rate;                     
    else                                                  
        *sampled_height = (height / *s_rate) + 1;

    determine_radii(*rb, *rs, rbt, rst);
    border = rbt[4] + dist_param * (*rs) + 1; 

    /* count the number of squares needed to cover the image */ 
    if (((*sampled_width - 2 * border) % (*square_width)) == 0) 
        *n_squares_x = (*sampled_width - 2 * border) / (*square_width);
    else
        *n_squares_x = ((*sampled_width - 2 * border) / (*square_width)) + 1;

    if (((*sampled_height - 2 * border) % (*square_width)) == 0) 
        *n_squares_y = (*sampled_height - 2 * border) / (*square_width);
    else
        *n_squares_y = ((*sampled_height - 2 * border) / (*square_width)) + 1;

    /* allocate memory for piece of the image */

    piece_height = 2 * border + *square_width;
  
    *image =  allocate_2d(piece_height, *sampled_width, item_size);

    /* allocate memory for marking coloured circles */  
    if (make_result_image == 1) {
        *circle_image = allocate_2d(piece_height, *sampled_width, item_size);
    }

    if ((*rowp = (void *) malloc(width * item_size)) == NULL) {
        fprintf(stderr, "Ethan: memory allocation failed, allocate_images\n");
        exit(ETHAN_MEMORY_ERROR);
    }
    *filtered_image = (void **) allocate_2d(*square_width, 
                                            *sampled_width - 2 * (*rb), 
                                            sizeof(int));
        
    /* allocate a table which can hold one peak per square */
    if ((*peaks = (struct peak *) malloc((*n_squares_x) * (*n_squares_y) 
         * sizeof(struct peak))) == NULL) {
        fprintf(stderr, "Ethan: memory allocation failed, allocate_images\n");
        exit(ETHAN_MEMORY_ERROR);
    }
}



/**************************************************************************/
/* mark a square                                                          */  
/**************************************************************************/
void draw_square(int r, int center_x, int center_y, unsigned char **image)
{
    int i;    

    for (i = center_x - r; i <= center_x + r; i++) {
        image[center_y - r][i] = DRAWING_VALUE;
        image[center_y + r][i] = DRAWING_VALUE; 
    }
    
    for (i = center_y - r; i <= center_y + r; i++) {
        image[i][center_x - r] = DRAWING_VALUE;
        image[i][center_x + r] = DRAWING_VALUE;
    }
}



/*****************************************************************************/
/* draw one square row to make a result image file                           */
/*****************************************************************************/
void draw_square_row_c(FILE *ofp, int width, int height, int rb, int rs, 
                       int border_height, int square_width, int n_squares_x, 
                       int n_squares_y, int squares_y, unsigned char **image, 
                       unsigned char **circle_image, 
                       struct peak *peaks, int draw_height, int s_rate)
{ 
    int i, y, peaks_start, peaks_end;
 
    /* during center refinement center may move to neighbour square row */
    peaks_start = (squares_y == 0) ? 0 : -1 * n_squares_x;
    peaks_end   = (squares_y == n_squares_y - 1) ? 
                                     n_squares_x : 2 * n_squares_x;


    /* draw squares around peaks of one square row */
    for (i = peaks_start; i < peaks_end; i++) {
        if (peaks[i].x != DELETED) { 

            /* calculate y coordinate within the piece */
            y = peaks[i].y - squares_y * square_width;
             
            /* if peak is in this square_row, draw squares */             
            if (y >= border_height && y < border_height + square_width) {
                draw_square(rs, peaks[i].x, y, circle_image);
                draw_square(rs + 1, peaks[i].x, y, circle_image);
                if (rs > 50)
                    draw_square(rs - 1, peaks[i].x, y, circle_image);
	    }
        }     
    } 
 
    /* write the top area of the piece into the file */       
    write_piece_ppm_c(ofp, image, circle_image, width, draw_height);
}    



/***************************************************************************/
/* go through the image to make a result image file                        */
/***************************************************************************/
void d_loop_square_rows_c(char *filename, FILE *ifp, 
                          FILE *ofp, int header_size,int width, int height, 
                          int sampled_width, int sampled_height, int s_rate, 
                          int item_size, unsigned char **piece_start, 
                          unsigned char **circle_image,  
                          int square_width, int n_squares_x, 
                          int n_squares_y, int rb, int rs, int border_height, 
                          struct peak *peaks)
{    
    int piece_height, rest_height, start_y, j;
    int square_row_size, diff_size;
    unsigned char **fill_start;
     
    piece_height = 2 * border_height + square_width;
    square_row_size = square_width * sampled_width * item_size;
    diff_size = piece_height * sampled_width - square_row_size * item_size;
     
    /* pointer to position in the piece where new square row is read */
    fill_start = piece_start + 2 * border_height;

    /* position in the original image from which reading the next piece */
    /* starts */ 
    start_y = 0; 

    if (n_squares_y > 1) {
      
        /* read the first square row and border areas above and below it */ 
        read_averaged_piece(filename, ifp, header_size, width, height, 
                            width, 0, 0, sampled_width, 
                            2 * border_height + square_width, s_rate,
                            item_size, piece_start);

        start_y = s_rate * (2 * border_height + square_width);        

        /* handle one square row at a time until two rows left */        
        for (j = 0; j < (n_squares_y - 2); j++) { 
            draw_square_row_c(ofp, sampled_width, piece_height,
                              rb, rs, border_height, square_width, 
                              n_squares_x, n_squares_y, j, 
                              piece_start, circle_image, 
                              peaks, square_width,s_rate);
            peaks += n_squares_x;

            memcpy(piece_start[0], piece_start[0] + square_row_size, 
                   diff_size);
            memcpy(circle_image[0], circle_image[0] + square_row_size, 
                   diff_size); 

            read_averaged_piece(filename, ifp, header_size, width, height, 
                               width, 0, start_y, sampled_width, 
                               square_width, s_rate, item_size, 
                               fill_start);

            memset((circle_image + 2 * border_height)[0], 0, square_row_size); 

            start_y += s_rate * square_width;      
        }
        draw_square_row_c(ofp, sampled_width, piece_height,
                          rb, rs, border_height, square_width, n_squares_x, 
                          n_squares_y, j, piece_start, circle_image, 
                          peaks, square_width, s_rate);   
        peaks += n_squares_x;
        j++;        

        memcpy(piece_start[0], piece_start[0] + square_row_size, 
               diff_size);
        memcpy(circle_image[0], circle_image[0] + square_row_size, 
               diff_size);

        /* handle the last. possibly not full square row */
        rest_height = sampled_height - 
                      (2 * border_height + (n_squares_y - 1) * square_width);

        read_averaged_piece(filename, ifp, header_size, width, height, 
                            width, 0, start_y, sampled_width, rest_height,
                            s_rate, item_size, fill_start);

        memset((circle_image + 2 * border_height)[0], 0, square_row_size);

        draw_square_row_c(ofp, sampled_width, 2 * border_height + rest_height,
                          rb, rs, border_height, square_width, n_squares_x, 
                          n_squares_y, j, piece_start, circle_image, peaks, 
                          2 * border_height + rest_height, s_rate);         

    } else {                                /* special case n_squares_y = 1 */ 
        read_averaged_piece(filename, ifp, header_size, width, height, 
                            width, 0, 0, sampled_width, sampled_height,
                            s_rate, item_size, piece_start);

        rest_height = sampled_height - 2 * border_height;       

        draw_square_row_c(ofp, sampled_width, sampled_height,
                          rb, rs, border_height, square_width, n_squares_x, 
                          n_squares_y, 0, piece_start, circle_image, peaks, 
                          sampled_height, s_rate);    
    }
}



/****************************************************************************/
/* filter one square row                                                    */
/****************************************************************************/
void filter_square_row_c(int width, int height, int rb, int rs, 
                         int square_width, int n_squares_x, int squares_y,
                         unsigned char **image, 
                         int **filtered_image, struct peak *peaks)
{   
    int i;    
  
    donut_filter_c(rb, rs, border, border,  width - 2 * border, 
                   height - 2 * border, image, filtered_image);
    find_peaks(square_width, 0, 0, width - 2 * border, 
               height - 2 * border, filtered_image, peaks);
    
    test_q(rb, rs, width, height, n_squares_x, image, peaks);    
   

    /* change peak position from position in the piece to position in the */
    /* whole image */
    for (i = 0; i < n_squares_x; i++) {
        if (peaks[i].x != DELETED) {
            peaks[i].y += border + squares_y * square_width;
        } 
    }
}



/***************************************************************************/
/* go through the image and process one square row at a time               */
/***************************************************************************/
void loop_square_rows_c(char *filename, FILE *ifp, 
                        int header_size,int width, int height, 
                        int border_height, int sampled_width, 
                        int sampled_height, int s_rate, int item_size, 
                        unsigned char **piece_start, int square_width, 
                        int n_squares_x, int n_squares_y, int rb, int rs, 
                        int **filtered_image, struct peak *peaks)
{ 
    int piece_height, rest_height, start_y, j;
    int square_row_size, diff_size;
    unsigned char **fill_start;
    int print_counter = 0, star_counter = 0;
     
    piece_height = 2 * border_height + square_width;
    square_row_size = square_width * sampled_width;
    diff_size = piece_height * sampled_width - square_row_size;
     
    /* pointer to position in the piece where new square row is read */
    fill_start = piece_start + 2 * border_height;

    /* position in the original image from which reading the next piece */
    /* starts */ 
    start_y = 0; 

    printf("Filtering %s: ", filename); 
    fflush(stdout);  
 
    if (n_squares_y > 1) {
      
        /* read the first square row and border areas above and below it */ 
        read_averaged_piece(filename, ifp, header_size, width, height, 
                            width, 0, 0, sampled_width, 
                            2 * border_height + square_width, s_rate,
                            item_size, piece_start);

        start_y = s_rate * (2 * border_height + square_width);        

        /* handle one square row at a time until two rows left */        
        for (j = 0; j < (n_squares_y - 2); j++) { 
            filter_square_row_c(sampled_width, piece_height,
                                rb, rs, square_width, n_squares_x, j,
                                piece_start, filtered_image, peaks);
            peaks += n_squares_x;

            memcpy(piece_start[0], piece_start[0] + square_row_size, 
                   diff_size);

            read_averaged_piece(filename, ifp, header_size, width, height, 
                                width, 0, start_y, sampled_width, 
                                square_width, s_rate, item_size, fill_start);

            start_y += s_rate * square_width;
            if ((star_counter < 10) &&
                (++print_counter  == ((n_squares_y - 2) / 10))) {
                printf("*");
                fflush(stdout);
                print_counter = 0;
                star_counter++;    
            }
        }
        printf("\n");
        filter_square_row_c(sampled_width, piece_height,
                            rb, rs, square_width, n_squares_x, j,
                            piece_start, filtered_image, peaks);
        j++;
        peaks += n_squares_x;
 
        memcpy(piece_start[0], piece_start[0] + square_row_size, 
               diff_size);

        /* handle the last. possibly not full square row */
        rest_height = sampled_height - 
                      (2 * border_height + (n_squares_y - 1) * square_width);

        read_averaged_piece(filename, ifp, header_size, width, height, 
                           width, 0, start_y, sampled_width, rest_height,
                           s_rate, item_size, fill_start);

        filter_square_row_c(sampled_width, 2 * border_height + rest_height,
                            rb, rs, square_width, n_squares_x, j,
                            piece_start, filtered_image, peaks);
        peaks += n_squares_x;

    } else {                                /* special case n_squares_y = 1 */ 
        read_averaged_piece(filename, ifp, header_size, width, height, 
                           width, 0, 0, sampled_width, sampled_height,
                           s_rate, item_size, piece_start);

        rest_height = sampled_height - 2 * border_height;       

        filter_square_row_c(sampled_width, sampled_height,
                            rb, rs, square_width, n_squares_x, 0,
                            piece_start, filtered_image, peaks);
        printf("\n");
    }
}



/****************************************************************************/
/* body of the virus detection                                              */
/****************************************************************************/
void find_viruses(int r, char *input_filename, char *coord_filename, 
                  char *dir_name, int s_rate)
{
    FILE *ifp, *pfp, *ofp;
    char output_filename[500];
    int width, height, item_size, header_size;
    struct peak *peaks;
    void **image, *rowp, **circle_image;
    void **filtered_image;
    int rb, rs, square_width, n_squares_x, n_squares_y;
    int sampled_width, sampled_height;
    int n, i, k; 
    char *suffix; 

    /* check that r is OK  */
    if (r < 3) {
        fprintf(stderr, "Ethan: radius is too small, find_viruses\n");
        exit(ETHAN_PARAMETER_ERROR);       
    }

    /* open the image file, pgm or mrc determined by name suffix*/
    if ((suffix = strstr(input_filename, ".pgm")) != NULL && 
        strlen(suffix) == 4) {
        ifp = open_pgm(input_filename, &width, &height, &item_size, 
                       &header_size);     
    } else {
        ifp = open_mrc(input_filename, &width, &height, &item_size, 
                       &header_size);
        if (item_size != sizeof(unsigned char)) {
            fprintf(stderr, "Ethan: error reading file, ");
            fprintf(stderr, "currently only byte files supported, ");
            exit(ETHAN_FORMAT_ERROR);       
        }
    }


    /* calculate the dimensions and allocate memory for image pieces */
    allocate_images(r, width, height, item_size, &sampled_width, 
                    &sampled_height, &s_rate, &rb, &rs, 
                    &square_width, &n_squares_x, &n_squares_y, &image, 
                    &filtered_image, &circle_image, &rowp, &peaks);

    if (r < s_rate) {
        fprintf(stderr, "Ethan: Radius is too small, find_viruses\n");
        exit(ETHAN_PARAMETER_ERROR);       
    }
    if (2 * (ring_width_param * r) + 1 > height ||
        2 * (ring_width_param * r) + 1 > height ||
        2 * rb + 1 > sampled_width || 2 * rb > sampled_height) {
        fprintf(stderr, "Ethan: radius or ring width parameter is too big,"); 
        fprintf(stderr, " find_viruses\n");
        exit(ETHAN_PARAMETER_ERROR);       
    }


    /* filter the image and find the peaks piece by piece */    
    loop_square_rows_c(input_filename, ifp, header_size, width, height, 
                       border, sampled_width, sampled_height, s_rate, 
                       item_size, (unsigned char **) image, square_width, 
                       n_squares_x, n_squares_y, rb, rs, 
                       (int **) filtered_image, peaks);  
		       
    n = n_squares_x * n_squares_y; 
    printf("Total number of squares is %d\n", n);
    k = 0;
    for (i = 0; i < n; i++) {
        if (peaks[i].x != -1) {
            k++;
	} 
    } 
    if (sector_test == 1)
        printf("Number of peaks after sector test is %d\n", k);
    else 
        printf("No sector test made\n");


    /* delete peaks that are too near other peaks, distance < rs */
    if (distance_test == 1) {
        edit_peaks(rs, n_squares_x, n_squares_y, peaks);
        k = 0;
        for (i = 0; i < n; i++) {
            if (peaks[i].x != -1) {
                k++;
	    } 
        }     
        printf("Number of peaks after first distance check is %d\n", k);
    } else {
        printf("No distance tests made\n");
    }

    /* delete peaks that are too small compared to the biggest */
    if (height_test == 1) {
        check_peaks(peaks, n);
        k = 0;
        for (i = 0; i < n; i++) {
            if (peaks[i].x != -1) {
                k++;
	    } 
        } 
        printf("Number of peaks after discarding too small ones is %d\n", k);
    } else {
        printf("No height test made\n");
    } 

    /* delete peaks that are too near other peaks, distance < 2 * rs */
    if (distance_test == 1) {
        delete_peaks(2 * rs - 0.1 * rs, n_squares_x, n_squares_y, peaks);
    } 

    /* try to find a more accurate centers and cut viruses if requested */
    if (refinement == 1) {
        printf("Refining centers\n");
        refine_and_cut(input_filename, dir_name, ifp, header_size, width, 
                       height, s_rate, ring_width_param * r, r, peaks, n);
    } else {
        printf("No center refinement made\n");
        for (i = 0; i < n; i++) {
            if (peaks[i].x != -1) {
                k++;
                peaks[i].x = peaks[i].x * s_rate;
                peaks[i].y = peaks[i].y * s_rate;
	    } 
        }              
    }

       
    /* store the virus coordinates */
    if ((pfp = fopen(coord_filename, "wb")) == NULL) {
        fprintf(stderr, "Ethan: error writing file %s\n", coord_filename);     
        exit(ETHAN_WRITE_ERROR); 
    }
    for (i = 0; i < n; i++)
        if (peaks[i].x != -1)
            fprintf(pfp, "%5d %5d\n", peaks[i].x, peaks[i].y);
    if (ferror(pfp)) {
        fprintf(stderr, "Ethan: error writing file %s\n", coord_filename);     
        exit(ETHAN_WRITE_ERROR); 
    }
    fclose(pfp);


    k = 0;
    for (i = 0; i < n; i++) {
        if (peaks[i].x != -1) {
            k++;
	} 
    } 

    printf("%d particles found from %s\n", k, input_filename);  

    /* make averaged image with viruses marked */
    if (make_result_image == 1) {
        for (i = 0; i < n; i++) {
            if (peaks[i].x != -1) {
                peaks[i].x = peaks[i].x / s_rate;
                peaks[i].y = peaks[i].y / s_rate;
	    } 
        } 

        strcpy(output_filename, coord_filename);
        strcat(output_filename, ".ppm"); 
        ofp = ppm_write_open(output_filename, sampled_width, sampled_height, 
                             UCHAR_MAX);
	
        d_loop_square_rows_c(input_filename, ifp, ofp, header_size, width, 
                             height, sampled_width, sampled_height, s_rate, 
                             item_size, (unsigned char **) image,  
                             (unsigned char **) circle_image, square_width, 
                             n_squares_x, n_squares_y, 
                             rb, rs, border, peaks);
        free_2d(circle_image);
        fclose(ofp);
   }

   free_2d(image);
   free_2d(filtered_image);
   free(rowp);
   fclose(ifp);        
}













































































