/*****************************************************************************
ETHAN 1.0 

Automated detection of spherical particles from electron micrographs
Copyright (C) 2000 Teemu Kivioja (Teemu.Kivioja@cs.Helsinki.FI)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.
*****************************************************************************/

#include "ethan.h"

#define update_edge_y(center_x, center_y, r, sum, edge) do {           \
        for (x = r; x > 0; x--) {                                      \
          sum += image[center_y + edge[x] + 1][center_x + x] +         \
	         image[center_y + edge[x] + 1][center_x - x] -         \
	         image[center_y - edge[x]][center_x + x] -             \
	         image[center_y - edge[x]][center_x - x];              \
        }	                                                       \
        sum += image[center_y + edge[0] + 1][center_x] -               \
	       image[center_y - edge[0]][center_x];                    \
    } while (0)  



#define update_edge_x(center_x, center_y, r, sum, edge) do {           \
        for (y = r; y > 0; y--) {                                      \
	  sum += image[center_y + y][center_x + edge[y]] -             \
                 image[center_y + y][center_x - edge[y] - 1] +         \
	         image[center_y - y][center_x + edge[y]] -             \
	         image[center_y - y][center_x - edge[y] - 1];          \
        }                                                              \
        sum += image[center_y][center_x + edge[0]] -                   \
	       image[center_y][center_x - edge[0] - 1];                \
    } while (0)      




/************************************************************************/
/* calculate the edge points of the filter circle (radius r) and store  */ 
/* them into table edge, only the first quarter is calculated           */
/************************************************************************/
int *calculate_edge(int r)
{	
  int p, x = 0, y = r;
  int *edge;
  
  if ((edge = malloc((r + 1)*sizeof(int))) == NULL) {
    fprintf(stderr, "Ethan: memory allocation failure, calculate_edge\n");
    exit(ETHAN_MEMORY_ERROR);
  }	
  edge[0] = r;
  edge[r] = 0;
  p = 1 - r;
  while (x < y) {
    if (p < 0) { 
      x ++;
    } else {
      x ++;
      y --;
    }
    if  (p < 0) {
      p = p + 2*x + 1;
    } else {
      p = p + 2*(x - y) + 1;
    }
    edge[x] = y;
    edge[y] = x;
  }  
  return edge;  
}



/***************************************************************************/
/* calculate the front edge and back edge of a circle                      */
/***************************************************************************/
void first_edge(int r, int center_y, int center_x, int *edge, 
                unsigned char **front_edge, unsigned char **back_edge, 
                unsigned char **image)
{
  int y;

  front_edge[r] = &image[center_y][center_x] + edge[0]; 
  back_edge[r]  = &image[center_y][center_x] - edge[0] - 1;                   
     
  y = r;
  do {
    front_edge[r + y] = &image[center_y + y][center_x] + edge[y];          
    front_edge[r - y] =	&image[center_y - y][center_x] + edge[y];         
    back_edge[r + y]  = &image[center_y + y][center_x] - edge[y] - 1;  
    back_edge[r - y]  = &image[center_y - y][center_x] - edge[y] - 1;          
  } while (--y > 0);                                                    
}



/***************************************************************************/
/*   calculate the sum of the pixel intensities in the circle              */
/*   and the area of the circle in pixels                                  */
/***************************************************************************/
int circle_sum_new(unsigned char **image, int r, int center_x, int center_y,
                   int *edge, int *area)
{
  int y, x;
  int sum = 0;


  *area = 0;
  for (y = r; y > 0; y--) {
    for (x = -edge[y]; x <= edge[y]; x++){
      sum += image[center_y + y][center_x + x] +
             image[center_y - y][center_x + x];
      *area = *area + 2;
    }
  }
  for (x = -edge[0]; x <= edge[0]; x++) {
    sum += image[center_y][center_x + x];   /* y = 0 */
    (*area)++;
  }
  return sum;
}



/**** arrays for octant check  *********************************************/

int ring_sum[8];
int sum[8];
float ring_mean[8];
float o_mean[8];
float mean[8];



/***************************************************************************/
/* Count the mean values of octants                                        */
/***************************************************************************/
void octant_means(int rb, int rs, int *edge_b, int *edge_s, int center_x, 
                  int center_y, unsigned char  **image)
{
    int y, x, start_x; 
    int i;
    float  n1 = 0, n2 = 0, rn1 = 0, rn2 = 0, rna, rnb;

    for (i = 0; i < 8; i++) {
        ring_sum[i] = 0;
        sum[i] = 0;
    }
   
    /* count the means of ring quadrants */    
    for (y = rb; y > rs; y--) {
        x = 1;
        while (x <= edge_b[y] && x < y) {
            ring_sum[1] += image[center_y + y][center_x + x];
            ring_sum[2] += image[center_y + y][center_x - x];
            ring_sum[5] += image[center_y - y][center_x - x];
            ring_sum[6] += image[center_y - y][center_x + x];
            rn1++;            
            x++;
        }
        for (x = y; x <= edge_b[y]; x++) {
            ring_sum[0] += image[center_y + y][center_x + x];
            ring_sum[3] += image[center_y + y][center_x - x];
            ring_sum[4] += image[center_y - y][center_x - x];
            ring_sum[7] += image[center_y - y][center_x + x];
            rn2++;
        }
    }
	
    for (y = rs; y > 0; y--) {
        x = edge_s[y];
        while (x <= edge_b[y] && x < y) {
            ring_sum[1] += image[center_y + y][center_x + x];
            ring_sum[2] += image[center_y + y][center_x - x];
            ring_sum[5] += image[center_y - y][center_x - x];
            ring_sum[6] += image[center_y - y][center_x + x];
            rn1++;            
            x++;
        }
        if (edge_s[y] > y)
            start_x = edge_s[y];
        else
            start_x = y; 
        for (x = start_x; x <= edge_b[y]; x++) {
            ring_sum[0] += image[center_y + y][center_x + x];
            ring_sum[3] += image[center_y + y][center_x - x];
            ring_sum[4] += image[center_y - y][center_x - x];
            ring_sum[7] += image[center_y - y][center_x + x];
            rn2++;
        }
    }	
    rna = 2 * rn1 + rn2;
    rnb = rn1 + 2 * rn2; 
  
    ring_mean[0] = (float) (ring_sum[7] + ring_sum[0] + ring_sum[1]) / rnb;
    ring_mean[1] = (float) (ring_sum[0] + ring_sum[1] + ring_sum[2]) / rna;
    ring_mean[2] = (float) (ring_sum[1] + ring_sum[2] + ring_sum[3]) / rna;
    ring_mean[3] = (float) (ring_sum[2] + ring_sum[3] + ring_sum[4]) / rnb;
    ring_mean[4] = (float) (ring_sum[3] + ring_sum[4] + ring_sum[5]) / rnb;
    ring_mean[5] = (float) (ring_sum[4] + ring_sum[5] + ring_sum[6]) / rna;
    ring_mean[6] = (float) (ring_sum[5] + ring_sum[6] + ring_sum[7]) / rna;
    ring_mean[7] = (float) (ring_sum[6] + ring_sum[7] + ring_sum[0]) / rnb;

    /* count the means of inner octants */
    for (y = rs; y > 0; y--) {
        x = 1;
        while (x < edge_s[y] && x < y) {
            sum[1] += image[center_y + y][center_x + x];
            sum[2] += image[center_y + y][center_x - x];
            sum[5] += image[center_y - y][center_x - x];
            sum[6] += image[center_y - y][center_x + x];
            x++;
            n1++;
        }
        for (x = y; x <= edge_s[y]; x++) {
            sum[0] += image[center_y + y][center_x + x];
            sum[3] += image[center_y + y][center_x - x];
            sum[4] += image[center_y - y][center_x - x];
            sum[7] += image[center_y - y][center_x + x];
            n2++;
        }  
    }	
    
    o_mean[0] = (float) (sum[0]) / (n2);
    o_mean[1] = (float) (sum[1]) / (n1);
    o_mean[2] = (float) (sum[2]) / (n1);
    o_mean[3] = (float) (sum[3]) / (n2);
    o_mean[4] = (float) (sum[4]) / (n2);
    o_mean[5] = (float) (sum[5]) / (n1);
    o_mean[6] = (float) (sum[6]) / (n1);
    o_mean[7] = (float) (sum[7]) / (n2);
     
}


/***************************************************************************/
/* test that none of the inner parts of sectors consists only of one color */
/***************************************************************************/
int black_area_check(int rs, int *edge_s, int center_x, int center_y, 
                     unsigned char  **image)
{
    int x, y;
    int black_area[8] = {YES, YES, YES, YES, YES, YES, YES, YES};
    int sector_value[8];
    int i;
    int first1256 = YES, first0347 = YES;

    for (y = rs; y > 0; y--) {
        x = 1;
        while (x < edge_s[y] && x < y) {
	    if (first1256 == YES) {
                first1256 = NO;
                sector_value[1] = image[center_y + y][center_x + x];
                sector_value[2] = image[center_y + y][center_x - x];
                sector_value[5] = image[center_y - y][center_x - x];
                sector_value[6] = image[center_y - y][center_x + x];
            } else {
                if (sector_value[1] != image[center_y + y][center_x + x])
                    black_area[1] = NO;
                if (sector_value[2] != image[center_y + y][center_x - x])
                    black_area[2] = NO;
                if (sector_value[5] != image[center_y - y][center_x - x])
                    black_area[5] = NO;
                if (sector_value[6] != image[center_y - y][center_x + x])
                    black_area[6] = NO;
            }
            x++;
        }
        for (x = y; x <= edge_s[y]; x++) {
	    if (first0347 == YES) {
                first0347 = NO;
                sector_value[0] = image[center_y + y][center_x + x];
                sector_value[3] = image[center_y + y][center_x - x];
                sector_value[4] = image[center_y - y][center_x - x];
                sector_value[7] = image[center_y - y][center_x + x];
            } else {
                if (sector_value[0] != image[center_y + y][center_x + x])
                    black_area[0] = NO;
                if (sector_value[3] != image[center_y + y][center_x - x])
                    black_area[3] = NO;
                if (sector_value[4] != image[center_y - y][center_x - x])
                    black_area[4] = NO;
                if (sector_value[7] != image[center_y - y][center_x + x])
                    black_area[7] = NO;
            }
        }  
    }	       
    for (i = 0; i < 8; i++) {
         if (black_area[i] == YES)
             return YES;
    }
    return NO;  
}



/***************************************************************************/
/*  check that an octant is darker than corresponding ring area and        */  
/*  and that difference between neighbour octants is smaller than          */
/*  difference between octant and the ring area and finally check that     */
/*  there are no black areas                                               */
/***************************************************************************/
int octant_check(int rb, int rs, int center_x, int center_y, int *n, 
                 unsigned char **image)
{
    int *edge_b;
    int *edge_s;


    edge_b = calculate_edge(rb);
    edge_s = calculate_edge(rs);  

    octant_means(rb, rs, edge_b, edge_s, center_x, center_y, image);

    free(edge_b);
    free(edge_s);      

    if (rs > 10)
        rs = rs / 2;
    edge_s = calculate_edge(rs);   


    if (o_mean[0] >= ring_mean[0] || o_mean[1] >= ring_mean[1] || 
        o_mean[2] >= ring_mean[2] || o_mean[3] >= ring_mean[3] ||
        o_mean[4] >= ring_mean[4] || o_mean[5] >= ring_mean[5] ||
        o_mean[6] >= ring_mean[6] || o_mean[7] >= ring_mean[7]) {
        free(edge_s);
        return NO;
    } else if (ring_mean[0] - o_mean[0] < fabs(o_mean[0] - o_mean[1]) || 
               ring_mean[1] - o_mean[1] < fabs(o_mean[1] - o_mean[2]) ||
               ring_mean[2] - o_mean[2] < fabs(o_mean[2] - o_mean[3]) ||
               ring_mean[3] - o_mean[3] < fabs(o_mean[3] - o_mean[4]) ||
               ring_mean[4] - o_mean[4] < fabs(o_mean[4] - o_mean[5]) ||
               ring_mean[5] - o_mean[5] < fabs(o_mean[5] - o_mean[6]) ||
               ring_mean[6] - o_mean[6] < fabs(o_mean[6] - o_mean[7]) ||
               ring_mean[7] - o_mean[7] < fabs(o_mean[7] - o_mean[0]) ||
               ring_mean[0] - o_mean[0] < fabs(o_mean[0] - o_mean[7]) || 
               ring_mean[1] - o_mean[1] < fabs(o_mean[1] - o_mean[0]) ||
               ring_mean[2] - o_mean[2] < fabs(o_mean[2] - o_mean[1]) ||
               ring_mean[3] - o_mean[3] < fabs(o_mean[3] - o_mean[2]) ||
               ring_mean[4] - o_mean[4] < fabs(o_mean[4] - o_mean[3]) ||
               ring_mean[5] - o_mean[5] < fabs(o_mean[5] - o_mean[4]) ||
               ring_mean[6] - o_mean[6] < fabs(o_mean[6] - o_mean[5]) ||
               ring_mean[7] - o_mean[7] < fabs(o_mean[7] - o_mean[6])) {
        free(edge_s);
        return NO;
    } else if (black_area_check(rs, edge_s, center_x, center_y, image) == 
               YES) {
        free(edge_s);
        return NO;
    } else {
        free(edge_s);
        return YES;  
    }
}



/**************************************************************************/
/* determine five different radii so that the best one can be found       */   
/* for each particle                                                      */
/**************************************************************************/
void determine_radii(int rb, int rs, int *rbt, int *rst)
{
    rst[0] = (1 - 2 * radius_coff) * rs;
    rst[1] = (1 -     radius_coff) * rs;
    rst[2] =                         rs;
    rst[3] = (1 +     radius_coff) * rs;
    rst[4] = (1 + 2 * radius_coff) * rs;

    rbt[0] = (1 - 2 * radius_coff) * rb;
    rbt[1] = (1 -     radius_coff) * rb;
    rbt[2] =                         rb;
    rbt[3] = (1 +     radius_coff) * rb;
    rbt[4] = (1 + 2 * radius_coff) * rb;

   /* check that the radii are different enough */
   if (rs - rst[1] < 2)
        rst[1] = rs - 2;
   if (rst[1] - rst[0] < 2)
        rst[0] = rst[1] - 2;

   if (rst[3] - rs < 2)
        rst[3] = rs + 2;
   if (rst[4] - rst[3] < 2)
        rst[4] = rst[3] + 2;   

   /* check that the rings are width enough */
   if (rbt[0] - rst[0] < 3)
       rbt[0] = rst[0] + 3;
   if (rbt[1] - rst[1] < 3)
       rbt[1] = rst[1] + 3;
   if (rbt[3] - rst[3] < 3)
       rbt[3] = rst[3] + 3;
   if (rbt[4] - rst[4] < 3)
       rbt[4] = rst[4] + 3;       
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void test_q(int rb, int rs, int width, int height, int n, 
            unsigned char **image, struct peak *peaks)  
{
    int i, j, area, result;
    int rbt[5], rst[5];
    int start_x, start_y, end_x, end_y, piece_width, piece_height;
    int dist, x, y, best;
    struct peak real_peak[5];

    /* determine an array of radii */
    determine_radii(rb, rs, rbt, rst);

    /* the amount of pixels the center is allowed move to one direction */
    dist = dist_param * rs;

    for (i = 0; i < n; i++) {
        if (peaks[i].x != DELETED) {
	    /* change the coordinates from filtered piece to piece */
	    /* with border */ 
 
            x = peaks[i].x + border;
            y = peaks[i].y + border;
    
            /* calculate the dimensions of refinement area */ 
            start_x = (x < dist + rbt[4]) ?  rbt[4] : x - dist;
            start_y = (y < dist + rbt[4]) ?  rbt[4] : y - dist;

            end_x   = (x + dist + rbt[4] > width - 1) ? 
                      width - rbt[4] - 1 : x + dist;
            end_y   = (y + dist + rbt[4] > height - 1) ? 
                      height - rbt[4] - 1 : y + dist;

            piece_width  = end_x - start_x + 1;
            piece_height = end_y - start_y + 1;

            /* try filters of different sizes */
            real_peak[0] = max_filter(rbt[0], rst[0], start_x, start_y, 
                                      piece_width, piece_height, image);
            real_peak[1] = max_filter(rbt[1], rst[1], start_x, start_y, 
                                      piece_width, piece_height, image);
            real_peak[2] = max_filter(rbt[2], rst[2], start_x, start_y, 
                                      piece_width, piece_height, image);

            real_peak[3] = max_filter(rbt[3], rst[3], start_x, start_y, 
                                      piece_width, piece_height, image);
            real_peak[4] = max_filter(rbt[4], rst[4], start_x, start_y, 
                                      piece_width, piece_height, image);


            /* choose the best result */                                       
            best = 0;
            for (j = 1; j < 5; j++)
                if (real_peak[best].height < real_peak[j].height) 
                    best = j;


            peaks[i].x = real_peak[best].x;
            peaks[i].y = real_peak[best].y;       
            peaks[i].height = real_peak[best].height;
            peaks[i].best = best;
        }              
    }

    for (i = 0; i < n; i++) {
        if (peaks[i].x != DELETED) {
	    if (sector_test == 1) {
                result = octant_check(rbt[peaks[i].best], rst[peaks[i].best], 
                                      peaks[i].x, peaks[i].y, &area, image);
            }
            if (sector_test == 1 && result == NO) {
	        peaks[i].x = DELETED;
            } else {
	        peaks[i].y = peaks[i].y - border;
	    }
        }
    }      

}



/***************************************************************************/
/* filter specified area using the ring filter                             */
/* fstart_x  and fstart_y: starting coordinates of the filtering           */
/* m width of area to be filtered (margins not included )                  */
/* n height of area to be filtered (margins not included)                  */
/***************************************************************************/
void donut_filter_c(int rb, int rs, int fstart_x, int fstart_y, int m, int n, 
                    unsigned char **image, int **filtered_image)    
{
    int *edge_b, *edge_s;           
    int sum_b, sum_s, sum_b1, sum_s1;
    int area_b, area_s;
    int center_x, center_y, x, y;   

    /* calculate edges of the circles */
    edge_b = calculate_edge(rb);
    edge_s = calculate_edge(rs);

    /* calculate the corner value */
    sum_b = circle_sum_new(image, rb, fstart_x, fstart_y, edge_b, &area_b);
    sum_s = circle_sum_new(image, rs, fstart_x, fstart_y, edge_s, &area_s);
    sum_b1 = sum_b;
    sum_s1 = sum_s;
    area_b -= area_s;
    filtered_image[0][0] = (int) (100 * 
                               ((float) (sum_b - sum_s) / (float) area_b - 
                                (float) sum_s / (float) area_s + 128.0)); 
  
    /* go iteratively through the whole image */
    for (center_y = fstart_y; center_y < fstart_y + n; center_y++) {
        for (center_x = fstart_x + 1; center_x < fstart_x + m; center_x++) {
            update_edge_x(center_x, center_y, rb, sum_b, edge_b);
            update_edge_x(center_x, center_y, rs, sum_s, edge_s); 
            filtered_image[center_y - fstart_y][center_x - fstart_x] = 
                           (int) (100.0 *  
	                         ((float) (sum_b - sum_s) / (float) area_b - 
                                 (float) sum_s / (float) area_s + 128.0));   
        }    
        if (center_y < fstart_y + n - 1) {
            update_edge_y(fstart_x, center_y, rb, sum_b1, edge_b); 
            update_edge_y(fstart_x, center_y, rs, sum_s1, edge_s);
            sum_b = sum_b1;
            sum_s = sum_s1;
            filtered_image[center_y - fstart_y + 1][0] = (int) (100 *
                          ( (float) (sum_b - sum_s) / (float) area_b - 
                          (float) sum_s / (float) area_s + 128.0));  
        }
    }   	
    free(edge_b);
    free(edge_s);
}



/*************************************************************************/
/* filter using maximum resolution                                       */
/*************************************************************************/
struct peak max_filter(int rb, int rs, int start_x, int start_y, 
                       int width, int height, unsigned char **piece) 
{
    int **filtered_piece;
    struct peak max;

    filtered_piece = (int **) allocate_2d(height, width, sizeof(int)); 
        
    donut_filter_c(rb, rs, start_x, start_y, width, height, piece, 
                   filtered_piece);

    max = find_max(0, 0, width, height, filtered_piece);
    max.x += start_x;
    max.y += start_y;

    free_2d((void **) filtered_piece);

    return max;
}












