/*****************************************************************************
ETHAN 1.0 

Automated detection of spherical particles from electron micrographs
Copyright (C) 2000 Teemu Kivioja (Teemu.Kivioja@cs.Helsinki.FI)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.
*****************************************************************************/

#include "ethan.h"

#define MRC_HEADER_SIZE 1024    /* size of the mrc file header */   
#define LITTLE 0                /* little endian               */
#define BIG 1                   /* big endian                  */

/****************************************************************************/
/*  swap the bytes of four byte integer                                     */
/****************************************************************************/
void swap(int *n)
{
   char *x;
   char temp;    

   x = (char *) n;

   temp = x[0];
   x[0] = x[3];
   x[3] = temp;
   
   temp = x[1];
   x[1] = x[2];
   x[2] = temp;
 
}



/***********************************************************************/
/* open mrc file named filename and read essential header information  */
/***********************************************************************/
FILE *open_mrc(char *filename, int *m, int *n, int *item_size, 
               int *header_size)
{
   int nz, mode; 
   int header[256];
   int endian = 1;
   int diff_endian = 0;
   FILE *ifp;  /* input data file */
 
      
   /* Attempt to open file */
   printf("Opening file %s\n", filename);
   if ((ifp = fopen(filename, "rb")) == NULL) {
      fprintf(stderr, "Ethan: error opening input file %s, open_mrc\n", 
              filename);
      exit(ETHAN_READ_ERROR);
   }
   
   /* Attempt to read the header */

   /* Read the header. */
   if (fread(header, 4, 256, ifp) != 256) {
      fprintf(stderr, "Error: error reading file %s, open_mrc\n", 
              filename);
      exit(ETHAN_READ_ERROR);  
   }
   /* check the endianess of the computer */
   if (*(char *)&endian == 1)
      endian = LITTLE; 

   /* we assume that we have different endianess in the file if            */
   /* the mode does not have value 0<mode<=4 OR mode is 0                  */
   /* and mapc is not 0<mapc<=3                                            */
   if ((header[3] < 0  || header[3] > 4)  ||
       (header[3] == 0 && (header[16] < 0 || header[16] > 3))) {
       diff_endian = 1;
       printf("different endianess in the file and machine\n");
   }
 
   *m  = header[0];
   if (diff_endian) 
      swap(m);
   printf("Width:  %d\n", *m);
   *n  = header[1];
   if (diff_endian) 
      swap(n);  
   printf("Height: %d\n", *n);
   nz = header[2];
   if (diff_endian) 
      swap(&nz);
   if (nz != 1) {
      fprintf(stderr, "Ethan: data file is not an image file %s, open_mrc\n",
             filename);
      exit(ETHAN_FORMAT_ERROR);
   }    
   mode = header[3];
   if (diff_endian) 
      swap(&mode);   
   if (mode != 0) {
      fprintf(stderr, "Ethan: data type not implemented %s, open_mrc\n",
              filename);
      exit(ETHAN_FORMAT_ERROR);
   }

   switch(mode) {
   case 0:               /* byte */
     *item_size = sizeof(unsigned char);
     break;
   case 1:               /* short */
     *item_size = sizeof(short);
     break;
   case 2:               /* float*/
     *item_size = sizeof(float);
     break;
   default:  /* other data types not implemented */
     fprintf(stderr, "Ethan: unknown mode %d in data file %s , open_mrc\n",
             mode, filename);
     exit(ETHAN_FORMAT_ERROR);
   }       
   *header_size = MRC_HEADER_SIZE;

   return ifp;  
}



/***********************************************************************/
/* open pgm file named filename and read essential header information   */
/***********************************************************************/
FILE *open_pgm(char *filename, int *width, int *height, int *item_size, 
               int *header_size)
                        
{
    FILE *f;
    unsigned char c;
    int sigma;

    printf("Opening file %s\n", filename);
    if (!(f = fopen(filename, "rb"))){
        fprintf(stderr, "Ethan: error opening input file %s, open_pgm\n", 
                filename);
        exit(ETHAN_READ_ERROR);
    }
    if ( fgetc( f ) != 'P' ) {
        fprintf(stderr, "Ethan: data file is not a pgm file %s, open_pgm\n",
                filename);
        exit(ETHAN_FORMAT_ERROR);
    }
    if ( fgetc( f ) != '5' ) {
        fprintf(stderr, "Ethan: data file is not a pgm file %s, open_pgm\n",
                filename);
        exit(ETHAN_FORMAT_ERROR);
    }
    fgetc( f ); /* newline */
    while(( c = fgetc( f )) == '#' ) 
        while( fgetc( f ) != '\n' );
    ungetc( c, f );

    fscanf( f, "%d %d %d", width, height, &sigma );
    fgetc( f );

    printf("Width:  %d\n", *width);
    printf("Height: %d\n", *height);

    *item_size = 1;
    *header_size = ftell(f);

    return f;
}



/****************************************************************************/
/* like read_sampled_piece but just sample, no averaging                    */
/****************************************************************************/
unsigned char **read_sampled_piece_c(char *filename, FILE *fp, int header_size,
                                     int width, int height, int piece_width_o, 
                                     int start_x, int start_y, 
                                     int sampled_width, int sampled_height, 
                                     int s_rate, int item_size,
                                     unsigned char **image, 
                                     unsigned char *rowp)
{
    int i, j, i_sampled, j_sampled, k = 0;

    /* i and k relate to original data, j and l to new sampled data */
    for (j = start_y, j_sampled = 0; j_sampled < sampled_height; 
         j = j + s_rate, j_sampled++) {
        fseek(fp, header_size + width*j + start_x, SEEK_SET);   
        if (fread(rowp, item_size, piece_width_o, fp) != piece_width_o) {
            fprintf(stderr, "Ethan: error reading data, %s, ", filename); 
            fprintf(stderr, "read_sampled_piece\n");
            free_2d((void **) image);
            free(rowp);
            exit(ETHAN_READ_ERROR);      
        } 
        for (i = 0, i_sampled = 0; i < piece_width_o; i += s_rate, 
             i_sampled++) {
            image[j_sampled][i_sampled] =  rowp[i];
            k++;
        }   
    }       
    return image;
}



/****************************************************************************/
/* read several lines and reduce them into one by averaging a square of     */
/* pixels into one                                                          */
/****************************************************************************/
void average_row(int width, int height, int sampled_width, int s_rate, 
                 unsigned char **rows, unsigned char *sampled_row)
{
    int i, j, k, l, num;    
    int **temp_rows;

    temp_rows = (int **) allocate_2d(s_rate, sampled_width, sizeof(int));  

    for (j = 0; j < height; j++) {    
        for (i = 0, l = 0; i < width; i += s_rate, l++) {
            for (k = i; k < i + s_rate; k++) {
                temp_rows[j][l] += rows[j][k];
	    }
        }            
    }
    for (l = 0; l < sampled_width; l++)
        for (j = height - 1; j > 0; j--)
            temp_rows[0][l] += temp_rows[j][l];

    num = height * s_rate; 
    for (i = 0; i < sampled_width; i++)
        sampled_row[i] = temp_rows[0][i] / num;

    free_2d((void **) temp_rows);
}



/****************************************************************************/
/* read piece of image and reduce its size by averaging square of pixels    */ 
/* into one                                                                 */
/****************************************************************************/
unsigned char **read_averaged_piece(char *filename, FILE *fp, int header_size,
                                    int width, int height, int piece_width_o, 
                                    int start_x, int start_y, 
                                    int sampled_width, int sampled_height, 
                                    int s_rate, int item_size,
                                    unsigned char **image)
{
    int j, j_sampled, k, left;
    unsigned char **rows;

    rows = (unsigned char **) allocate_2d(s_rate, width, item_size);    

    for (j = start_y, j_sampled = 0; j_sampled < sampled_height - 1; 
         j = j + s_rate, j_sampled++) {

        for (k = 0; k < s_rate; k++) {
           fseek(fp, header_size + width * (j + k) + start_x, SEEK_SET);  
           if (fread(rows[k], item_size, piece_width_o, fp) != 
               piece_width_o) {
                fprintf(stderr, "Ethan: error reading file, %s, ", filename); 
                fprintf(stderr, "read_sampled_piece\n");
                exit(ETHAN_READ_ERROR);
           }                 
        }
        average_row(width, s_rate, sampled_width, s_rate, rows, 
                    image[j_sampled]);
    }       
 
    if (height - start_y < s_rate * sampled_height) 
        left = height - (start_y + s_rate * (sampled_height - 1));
    else
        left = s_rate;
                
    for (k = 0; k < left; k++) {
        fseek(fp, header_size + width * (j + k) + start_x, SEEK_SET);  
        if (fread(rows[k], item_size, piece_width_o, fp) != 
            piece_width_o) {
            fprintf(stderr, "Error: error reading file, %s, ", filename); 
            fprintf(stderr, "read_sampled_piece\n");
            exit(ETHAN_READ_ERROR);
        }                 
    }
    average_row(width, left, sampled_width, s_rate, rows, 
                image[j_sampled]);

    free_2d((void **) rows);
    return image;
}



/***************************************************************************/
/* open mrc file for writing and write the header                          */
/***************************************************************************/
FILE *mrc_write_open(char *filename, int nx, int ny, int nz)
{
    int header[256];
    int i;
    FILE *fp;

    /* most values are 0 */
    for (i = 4; i < 256; i++)
        header[i] = 0;

    header[0] = nx;
    header[1] = ny;
    header[2] = nz;
    header[3] = 0;
    
    header[16] = 1; /* map */
    header[17] = 2; 
    header[18] = 3;

    header[55] = 1; /* nlabl */
   

    if (!(fp = fopen( filename, "wb"))){
        fprintf(stderr, "Ethan: Error opening file %s, mrc_write_open.\n", 
                filename );
	exit(ETHAN_WRITE_ERROR);
    }

    /* Write header to disk. */
    if (fwrite(header, 1, 1024, fp) != 1024) {
        fprintf(stderr, "Ethan: Error writing file %s, mrc_write_open.\n", 
                filename );
        exit(ETHAN_WRITE_ERROR);
    }
    return fp;

}


   
/************************************************************************/
/* open a ppm file for writing                                          */
/************************************************************************/
FILE *ppm_write_open(char *filename, int w, int h, int sigma)
{
    FILE *fp;

    fp = fopen( filename, "wb" );

    fprintf(fp, "P6\n");
    fprintf(fp, "%d %d\n%d\n", w, h, sigma );
       
    return fp;
}



/************************************************************************/
/* write piece of the result image file                                 */
/************************************************************************/
void write_piece_ppm_c(FILE *fp, unsigned char **image, 
                       unsigned char **circle_image, int width, int height)
{       
    int i, j;

    for (j = 0; j < height; j++)
        for (i = 0; i < width; i++) {
            if (circle_image[j][i] == DRAWING_VALUE) {  
                fputc(UCHAR_MAX, fp);
                fputc(0, fp);
                fputc(0, fp);
            } else { 
                fputc(image[j][i], fp);
                fputc(image[j][i], fp);
                fputc(image[j][i], fp);
	    }
        }
}















